const AppDisplay   = imports.ui.appDisplay;
const AppFavorites = imports.ui.appFavorites;
const AppSystem    = imports.gi.Shell.AppSystem.get_default();
const GMenu        = imports.gi.GMenu;
const Mainloop     = imports.mainloop;
const Me           = imports.misc.extensionUtils.getCurrentExtension();
const ViewSelector = imports.ui.main.overview._viewSelector;

const Prefs = Me.imports.prefs;

const SEARCH_CATEGORY = 'b88f323f-bff3-4106-8c21-de6974f38ede';

let providers = [];
let undos     = [];

function init() {
    let jindex = let (
        menus = Prefs.getMenus(),
        x = -1
    ) function(dir) {
        for (let i = 0; i < menus.length; ++i) {
            for (let j = 0; j< menus[i].length; ++j) {
                if (menus[i][j] == dir.get_menu_id()) {
                    return [i, j];
                }
            }
        }
        return [menus.length, ++x];
    };

    let hidden = let (
        hidden = Prefs.getHiddenApps().concat(Prefs.getMovedApps()).map(function(app, i, a) {
            return AppSystem.lookup_app(app);
        }).filter(function(app, i, a) {
            return typeof app != 'undefined';
        })
    ) function(app) {
        return hidden.indexOf(app) >= 0 || app.get_app_info().get_nodisplay();
    };

    function walk(dir) {
        if (dir.get_is_nodisplay()) {
            return [];
        }
        let dirs = [];
        let apps = [];
        let iter = dir.iter();
        for (let t = iter.next(); t != GMenu.TreeItemType.INVALID; t = iter.next()) {
            switch (t) {
            case GMenu.TreeItemType.DIRECTORY:
                dirs = dirs.concat(walk(iter.get_directory()));
                break;
            case GMenu.TreeItemType.ENTRY:
                let app = AppSystem.lookup_app_by_tree_entry(iter.get_entry());
                if (!hidden(app)) {
                    apps.push(app);
                }
                break;
            }
        }
        if (apps.length > 0) {
            let [i, j] = jindex(dir);
            dirs.push({
                i    : i,
                j    : j,
                dir  : dir,
                apps : apps
            });
        }
        return dirs;
    }

    function title(dirs) {
        let title = dirs.map(function(dir, i, a) {
            return dir.get_name().replace(/\s*&/g, ',');
        }).join(', ');

        let c = title.lastIndexOf(',');
        if (c > 0 && c < title.length - 2) {
            return title.substring(0, c) + ' &' + title.substring(c + 1);
        }
        return title;
    };

    function appSearchProvider(title, apps) {
        let provider = new AppDisplay.AppSearchProvider();
        provider.title = title;
        provider.getSubsearchResultSet = function(previousResults, terms) {
            provider.getInitialResultSet(terms);
        };
        provider.getInitialResultSet = function(terms) {
	        if (terms == SEARCH_CATEGORY) {
                provider.searchSystem.pushResults(provider, apps);
	        } else {
	            provider.searchSystem.pushResults(provider, []);
	        }
        };
        return provider;
    }

    providers = walk(AppSystem.get_tree().get_root_directory()).sort(function(a, b) {
        if (a.i < b.i || a.i == b.i && a.j < b.j) {
            return -1;
        }
        if (a.i > b.i || a.i == b.i && a.j > b.j) {
            return  1;
        }
        return 0;
    }).reduce(function(groups, x, i, a) {
        if (typeof groups[x.i] == 'undefined') {
            groups[x.i] = {
                dirs : [],
                apps : []
            };
        }
        groups[x.i].dirs.push(x.dir);
        groups[x.i].apps = groups[x.i].apps.concat(x.apps);
        return groups;
    }, []).map(function(group, i, a) {
        return {
            enable : function() {
                function appsSort(a, b) {
		            let aFav = AppFavorites.getAppFavorites().isFavorite(a.get_id());
		            let bFav = AppFavorites.getAppFavorites().isFavorite(b.get_id());
                    
		            return aFav != bFav ? (aFav ? -1 : 1) : a.compare_by_name(b);
                }

                let provider = appSearchProvider(title(group.dirs), group.apps.sort(appsSort));

                ViewSelector.addSearchProvider(provider);
                ViewSelector._searchResults._providerMeta.filter(function(meta, i, a) {
                    return meta.provider === provider;
                }).forEach(function(meta, i, a) {
                    meta.resultDisplay._grid._rowLimit = 10;
                });

                undos.push(function() {
                    ViewSelector.removeSearchProvider(provider);
                });

                undos.push(Prefs.connect(AppFavorites.getAppFavorites(), 'changed', function() {
                    if (group.apps.slice(0) != group.apps.sort(appsSort)
                        && provider.searchSystem.getTerms() == SEARCH_CATEGORY) {
                        Mainloop.timeout_add(Prefs.getWorkaroundDelay(), function() {
                            provider.searchSystem.updateSearch(SEARCH_CATEGORY);
                        });
                    }
                }));
            }
        };
    });
}

function enable() {
    function reset() {
        let redisplay = ViewSelector._searchSystem.getTerms() == SEARCH_CATEGORY;

        disable();
        init();
        enable();

        if (redisplay) {
            ViewSelector._searchSystem.updateSearch(SEARCH_CATEGORY);
        }
    }
    providers.forEach(function(provider, i, a) {
        provider.enable();
    });
    undos.push(function() {
        ViewSelector._searchResults.reset();
    });

    undos.push(Prefs.onAppsChange(reset));
    undos.push(Prefs.connect(AppSystem, 'installed-changed', reset));
}

function disable() {
    while(undos.length > 0) {
        undos.pop()();
    }
}
